/**
  Lighting demo application.
  
  Orion Sky Lawlor, olawlor@acm.org, 2008-04-08 (Public Domain)
*/
#include <GL/glew.h> /*<- for gl...ARB extentions.  Must call glewInit after glutCreateWindow! */
#include <iostream>
#include <GL/glut.h>
#include <math.h>
#include "osl/vec4.h" /*<- handy little vector class */
#include "osl/mat4.h"
using osl::mat4;
/* Convert a struct with .x, .y, .z fields into three arguments */
#define VEC3_TO_XYZ(v) (v).x,(v).y,(v).z
#ifndef M_PI
#define M_PI 3.14159265
#endif

#include "ogl/glsl.h" /* Orion's tiny GLSL library */
#include "ogl/glsl.cpp"

#include "ogl/minicam.h"

#include "GL/glui.h" /* OpenGL User Interface library (for making a GUI) */

void oglExit(const char *what,const char *why) {
	printf(what,why);
	exit(1);
}
int gi_level=5; /* power of 2 pixel count for little global illumination views */

int gi_self=1; /* show object while computing its own GI views */
int show_sky=1; /* show sky */
int show_ground=1; /* show ground plane */
int show_noise=1; /* show noise texture */
int show_light=1; /* show light source */

float gi_fov=130.0; /* field of view, in degrees, for little GI views */
float obj_reflectance=0.8;
float obj_scale=1.0;
int move_camera=1, show_gi=0;
mat4 model_gui=mat4(1.0);

int cos_falloff=1;

mat4 lightMat(1.0);
float lightsize=4.0;
int light_antialias=1;
// These are the corners of our light source.
//  They get transformed by lightMat and lightsize.
vec3 lightCorners[3]={vec3(-1,0,1), vec3(-1,-1,1), vec3(0,0,1) };

/******************* Object, read from .obj file ***************/
class obj_mesh {
public:
	obj_mesh(const char *fName);
	
	void draw(void);
	
	void gi(const osl::mat4 &to_World);
private:
	struct myVertex {
		vec3 xyz; /* location */
		vec3 color; /* GI-derived color */
		myVertex(const vec3 &v) :xyz(v), color(0,1,0) {}
	};
	std::vector<myVertex> vtx; /* 3D location of vertices (0-based) */
	GLuint vb, eb; /* GPU vertex buffer object, element buffer object */
	void vb_upload(void); /* copy vtx array data into vb */
	
	std::vector<vec3> vtx_norm; /* GI vertex normals */
	
	/* By normal index */
	std::vector<vec3> norm; /* surface normal (unit length) */
	
	struct myFace {
		int v[3]; /* 1-based vertex index */
		int n[3]; /* 1-based normal vector index */
	};
	std::vector<myFace> tri;
};

obj_mesh::obj_mesh(const char *fName)
{
	FILE *f=fopen(fName,"r");
	char line[1000], stuff[100];
	while (NULL!=fgets(line,999,f)) {
		switch (line[0]) {
		case 'v': { /* Vertex line, like "v 0.1 0.7 0.9" */
			double x,y,z;
			if (4!=sscanf(line,"%99s %lf%lf%lf",stuff,&x,&y,&z)) 
				oglExit("Unparsable v-line '%s'\n",line);
			if (stuff==std::string("v"))
				vtx.push_back(vec3(x,y,z));
			else if (stuff==std::string("vn"))
				norm.push_back(normalize(vec3(x,y,z)));
			else
				oglExit("Unknown v-line '%s'\n",line);
			break;
		}
		case 'f': { /* Face line, like "f 1 2 3 4" or "f 1//1 2//2 3//3 4//4". */
			myFace q;
			for (int i=0;i<3;i++) q.n[i]=0;
			/* FIXME: this is hideous coding, right here... */
			/* Tri face, with normals */
			if (7==sscanf(line,"%99s %d//%d  %d//%d  %d//%d ",stuff,
				&q.v[0],&q.n[0], 
				&q.v[1],&q.n[1], 
				&q.v[2],&q.n[2])
			)
			{
				vtx_norm.resize(vtx.size());
				for (int i=0;i<3;i++)
					vtx_norm[q.v[i]-1]=norm[q.n[i]-1];
				tri.push_back(q);
			}
			/* Tri face, without normals */
			else if (4==sscanf(line,"%99s %d  %d  %d ",stuff,
					&q.v[0], 
					&q.v[1], 
					&q.v[2])
			)
				tri.push_back(q);
			else /* all parse attempts failed! */
				oglExit("Unknown f-line '%s'\n",line);
			break;
		}
		
		default: break; /* ignore unrecognized lines */
		};
	}
	fclose(f);
	
	/* Make triangle index buffer and upload to graphics card */
	std::vector<unsigned int> tri_idx;
	for (unsigned int t=0;t<tri.size();t++) {
		for (int i=0;i<3;i++) 
			tri_idx.push_back(tri[t].v[i]-1);
	}
	glGenBuffersARB(1,&eb); 
	glBindBufferARB(GL_ELEMENT_ARRAY_BUFFER_ARB,eb);
	glBufferDataARB(GL_ELEMENT_ARRAY_BUFFER_ARB,sizeof(tri_idx[0])*tri_idx.size(),
		&tri_idx[0],GL_STATIC_DRAW_ARB);
	glBindBufferARB(GL_ELEMENT_ARRAY_BUFFER_ARB,0);
	
	/* Make vertex buffer, and upload to graphics card */
	glGenBuffersARB(1,&vb); 
	vb_upload();
}
void obj_mesh::vb_upload(void) 
{ /* copy vtx array data into vb */
	glBindBufferARB(GL_ARRAY_BUFFER_ARB,vb);
	glBufferDataARB(GL_ARRAY_BUFFER_ARB,sizeof(vtx[0])*vtx.size(),
		&vtx[0],GL_STREAM_DRAW_ARB);
	glEnableClientState(GL_VERTEX_ARRAY);
	glEnableClientState(GL_COLOR_ARRAY);
	glVertexPointer(3,GL_FLOAT,sizeof(vtx[0]), (void *)0); /* myVertex.xyz is first thing in struct */
	glColorPointer (3,GL_FLOAT,sizeof(vtx[0]), (void *)12); /* myVertex.color starts 12 bytes after struct start */
	glBindBufferARB(GL_ARRAY_BUFFER_ARB,0);
}
void obj_mesh::draw(void)
{
	/* Turn everything on */
	glBindBufferARB(GL_ELEMENT_ARRAY_BUFFER_ARB,eb);
	glBindBufferARB(GL_ARRAY_BUFFER_ARB,vb);
	/* Draw our (GPU) triangles with (GPU) vertices */
	glDrawElements(GL_TRIANGLES,tri.size()*3,GL_UNSIGNED_INT,0);
	
	/* Turn everything off afterwards */
	glBindBufferARB(GL_ARRAY_BUFFER_ARB,0);
	glBindBufferARB(GL_ELEMENT_ARRAY_BUFFER_ARB,0);
}
/*********************** App Code **********************/

/* Orthonormal coordinate frame */
/* Time it took to draw the last frame (for updating physics) */
float seconds_per_frame=0.01;

/* Nice handy timer: returns seconds */
double nice_timer(void) {return glutGet(GLUT_ELAPSED_TIME)*1.0e-3;}

/* Return a random float on [-1,+1] */
float rand_float(void) {
	return (rand()&0xffff)*(1.0/0xffff)*2.0-1.0;
}

/* Return a random two-vector on [-1,+1] */
vec3 rand_vec2(void) {
	return vec3(rand_float(),rand_float(),0.0);
}
/* Return the fractional part of this vector */
vec3 fract(const vec3 &v) {
	return v-vec3(floor(v.x),floor(v.y),floor(v.z));
}

/**
 Stores an entire OpenGL scene.
*/
class the_world {
public:
	the_world();
	
	/* Precompute values needed for a given frame.  May trash matrices! */
	void setup(void);
	
	/* Render world to the current modelview and projection matrices */
	void render(bool for_camera);
private:
	int lit_sz; /* resolution of plane-lighting */
	
	obj_mesh model; /* our main model */
	mat4 modelM; /* model-to-world coordinate shift */
};


/// Make a new empty OpenGL texture with this size and internal format.
GLuint make_tex(int w,int h,GLenum texFormat) {
	GLuint tex;
	glGenTextures(1, &tex);
	glBindTexture(GL_TEXTURE_2D, tex);
	/* turn on mipmapping */
	GLenum filt=GL_LINEAR_MIPMAP_LINEAR;
	if (texFormat==GL_RGBA_FLOAT16_ATI) filt=GL_NEAREST; /* or else nVidia 6200 fails below */
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, filt);
	for (int level=0;0<(w>>level);level++) // add all mipmap levels
		glTexImage2D(GL_TEXTURE_2D, level, texFormat, 
			w>>level, h>>level, 0, GL_RGBA, GL_UNSIGNED_BYTE, NULL);
	return tex;
}

/// Make a texture to represent the cosine amplitude fall-off for this field of view:
GLuint make_cosine_tex(int w,int h,GLenum texFormat,float fov,float &avg)
{
	float f=tan(fov*(M_PI/180.0)*0.5); // see gluPerspective
	
	std::vector<float> falloff(w*h);
	avg=0.0;
	for (int y=0;y<h;y++)
	for (int x=0;x<w;x++) {
		float xG=x*(2.0/(w-1))-1.0; // OpenGL x, from -1 to +1
		float yG=y*(2.0/(h-1))-1.0;
		vec3 lookdir=normalize(vec3(f*xG,f*yG,1.0));
		float c=pow(lookdir.z,2.0f); // == dot(lookdir,vec3(0.0,0.0,1.0))^2 == cos(angle to look direction)^2
		avg+=c;
		falloff[y*w+x]=c;
	}
	avg=avg/(w*h);
	
	GLuint tex=make_tex(w,h,texFormat);
	glTexImage2D(GL_TEXTURE_2D,0,texFormat,w,h,0,
		GL_LUMINANCE,GL_FLOAT,&falloff[0]);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	return tex;
}

the_world::the_world()
	:model("screamer.obj")
{

}

vec3 noiseTexCoord(vec3 pos) {
	return 0.05*(pos+vec3(0.0,0.0,1.0*glutGet(GLUT_ELAPSED_TIME)*0.001));
}

/* Render the world, either to a GI view or to the camera.  
   Assumes matrices are already set up. */
void the_world::render(bool from_camera) {
/* Clear the sky */
	glClear(GL_COLOR_BUFFER_BIT + GL_DEPTH_BUFFER_BIT);
	
	float z=0.0;
	if (show_ground)
	{
		/* Draw the ground */
		glColor3f(0.3,0.2,0.1);
		glBegin(GL_TRIANGLE_FAN);
		glVertex3f(-20.0,-10.0,z);
		glVertex3f(+20.0,-10.0,z);
		glVertex3f(0.0,+20.0,z);
		glEnd();
	}
	
	if (show_light) 
	{
		/* Draw the light source */
		glColor3f(1.0,1.0,1.0);
		glBegin(GL_TRIANGLES);
		for (int l=0;l<3;l++)
			glVertex3fv(lightMat*vec4(lightsize*lightCorners[l],1.0));
		glEnd();
		if (light_antialias) {
			// outline with smooth lines, for antialiasing
			glEnable(GL_LINE_SMOOTH);
			glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
			glEnable(GL_BLEND);
			glBegin(GL_LINE_LOOP);
			for (int l=0;l<3;l++)
				glVertex3fv(lightMat*vec4(lightsize*lightCorners[l],1.0));
			glEnd();
			glDisable(GL_BLEND);
		}
	}


	if (show_noise)
	{
		/* Draw the static box */
		static GLuint noiseTex=0;
		if (noiseTex==0) 
		{ /* only need to build and upload one time; it stays set up. */
			glGenTextures(1,&noiseTex);
			glBindTexture(GL_TEXTURE_3D,noiseTex);
			int sz=16; /* RGBA 32 x 32 x 32 */
			float *noise=new float[4*sz*sz*sz];
			for (int z=0;z<sz;z++) 
			for (int y=0;y<sz;y++) 
			for (int x=0;x<sz;x++) 
			for (int chan=0;chan<4;chan++) {
				float r=(rand()&0xfff)*(1.0/0xfff);
				noise[chan+4*(x+sz*(y+sz*z))]=r;
			}
			glTexImage3D(GL_TEXTURE_3D,0,GL_RGBA8,
				sz,sz,sz,0,
				GL_RGBA,GL_FLOAT,noise);
			glTexParameteri(GL_TEXTURE_3D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
			glTexParameteri(GL_TEXTURE_3D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
			delete[] noise;
		}
		glBindTexture(GL_TEXTURE_3D,noiseTex);
		glEnable(GL_TEXTURE_3D);

		glColor3f(1.0,1.0,1.0);
		glBegin(GL_TRIANGLE_FAN);
		float x=3.3;
		vec3 p;
		p=vec3(x,-5.0,z);      glTexCoord3fv(noiseTexCoord(p)); glVertex3fv(p);
		p=vec3(x, 0.0,z+10.0); glTexCoord3fv(noiseTexCoord(p)); glVertex3fv(p);
		p=vec3(x,+5.0,z);      glTexCoord3fv(noiseTexCoord(p)); glVertex3fv(p);
		glEnd();
		glDisable(GL_TEXTURE_3D);
	}


	if (!from_camera && !gi_self) return; /*<- don't draw object in its own images */
	
	/* Draw the model, with GI vertex colors */
	glLoadMatrixf(&modelM[0][0]);
	model.draw();
	glLoadIdentity();
}

the_world *world=0; /* world model (can't initialize it until OpenGL is set up!) */

/* Do global illumination calculation for this mesh, 
  oriented according to this matrix. */
void obj_mesh::gi(const mat4 &toWorld)
{
	int gi_res=1<<gi_level; // pixels per side of each vertex's view
	static int sz=0; // square per-vertex texture, vertices per side
	if (sz==0) {
		sz=1;
		while (sz*sz<(int)vtx.size()) sz*=2;
	}
	static GLuint rgbTex=0, rgbTexLevel=0;
	static GLuint rgbFBO=0;
	if (rgbTex==0 || rgbTexLevel!=(GLuint)gi_level) {
		rgbTexLevel=gi_level;
		rgbTex=make_tex(sz*gi_res,sz*gi_res,GL_RGBA8);
		glGenFramebuffersEXT(1,&rgbFBO);
		// Need a Z buffer for GL_DEPTH_TEST to work...
		GLuint depthTex=make_tex(sz*gi_res,sz*gi_res,GL_DEPTH_COMPONENT24);
		glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, 
			GL_DEPTH_ATTACHMENT_EXT, GL_TEXTURE_2D, depthTex, 0);
	}

// Draw the world into little per-vertex views, tiled into our rgb texture
	if (!show_gi) glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, rgbFBO);
	glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, 
		GL_COLOR_ATTACHMENT0_EXT, GL_TEXTURE_2D, rgbTex, 0);
	
	/* Figure out how much light is coming in at each mesh vertex */
	glEnable(GL_SCISSOR_TEST);
	int vx=-1, vy=0; // location of v's tile
	for (unsigned int v=0;v<vtx.size();v++) {
		vx++; 
		if (vx>=sz) {
			vx=0; 
			vy++;
			if (vy>=sz) {
				printf("Error!  Too many vertices!\n");
				exit(1);
			}
		}
		int render_x=vx*gi_res,render_y=vy*gi_res;
		
	/* Set up camera pointing out from this vertex */
		vec3 p=vec3(toWorld*vec4(vtx[v].xyz,1.0));
		vec3 n=vec3(toWorld*vec4(vtx_norm[v],0.0));
		vec3 y=normalize(cross(vec3(0,1,0),n));

		glMatrixMode(GL_PROJECTION); /* set up projection matrix */
		glLoadIdentity(); /* reset matrix at start of frame */
		gluPerspective(gi_fov,1.0,0.01,1000.0);
		gluLookAt(
			VEC3_TO_XYZ(p), /* camera */
			VEC3_TO_XYZ(p+n), /* looking at this point */
			VEC3_TO_XYZ(y) /* up vector */
		);
		glMatrixMode(GL_MODELVIEW); /* set up model and view matrix */
		glLoadIdentity(); /* reset matrix at start of frame */

	/* Set up framebuffer to render our small region */
		glViewport(render_x,render_y,gi_res,gi_res);
		glScissor(render_x,render_y,gi_res,gi_res);

	/* Render everything */
		world->render(false);
	}
	
	float cosAvg=1.0;
	if (cos_falloff) {
		/* Multiply in the cosine fall-off texture */
		static float cosAvgLast=0.0, cosFOV=0.0;
		static GLuint cosTex=0;
		if (cosFOV!=gi_fov) {
			cosTex=make_cosine_tex(gi_res,gi_res,GL_LUMINANCE8,gi_fov,cosAvgLast);
			cosFOV=gi_fov;
		}
		cosAvg*=cosAvgLast;
		glViewport(0,0,sz*gi_res,sz*gi_res);
		glScissor(0,0,sz*gi_res,sz*gi_res);
		glBlendFunc(GL_DST_COLOR,GL_ZERO);// multiplicative blending
		glEnable(GL_BLEND);
		glDisable(GL_DEPTH_TEST);
		glBindTexture(GL_TEXTURE_2D,cosTex);
		glEnable(GL_TEXTURE_2D);
		glMatrixMode(GL_PROJECTION);
		glLoadIdentity();
		glMatrixMode(GL_MODELVIEW);
		glLoadIdentity();

		glColor3f(1,1,1);
		glBegin(GL_TRIANGLE_FAN);
			glTexCoord2f(0,0); glVertex2f(-1,-1);
			glTexCoord2f(sz,0); glVertex2f(+1,-1);
			glTexCoord2f(sz,sz); glVertex2f(+1,+1);
			glTexCoord2f(0,sz); glVertex2f(-1,+1);
		glEnd();

		glDisable(GL_TEXTURE_2D);
		glDisable(GL_BLEND);
		glEnable(GL_DEPTH_TEST);
	}
	
	if (show_gi) return;
	
/* Rendered the world, now read back the averaged-down color from mipmap */
	glBindTexture(GL_TEXTURE_2D,rgbTex);
	glEnable(GL_TEXTURE_2D); /*<- makes glGenerateMipmapEXT work on ATI! */
	glGenerateMipmapEXT(GL_TEXTURE_2D);
	
	glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, // for read
		GL_COLOR_ATTACHMENT0_EXT, GL_TEXTURE_2D, rgbTex, gi_level); // <- mipmap!
	static std::vector<unsigned char> pix(sz*sz*3);
	glPixelStorei(GL_PACK_ALIGNMENT,1); /* allow byte alignment */
	glReadPixels(0,0,sz,sz,GL_RGB,GL_UNSIGNED_BYTE,&pix[0]);
	for (unsigned int v=0;v<vtx.size();v++) {
		vec3 light=vec3(pix[3*v+0],pix[3*v+1],pix[3*v+2]);
		light*=(1.0/(255*cosAvg));
		vec3 reflectance(obj_reflectance);/* this vertex's reflectance */
		vtx[v].color=light*reflectance; 
	}
	// Back to the normal framebuffer
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);
	glDisable(GL_TEXTURE_2D);
	
	/* Now that we know mesh colors, update the GPU's copy of the mesh */
	vb_upload();
}

/* Precompute values needed for a given frame */
void the_world::setup(void)
{
	/* Capture the model matrix */
	glMatrixMode(GL_MODELVIEW); /* set up model and view matrix */
	glLoadIdentity(); /* reset matrix at start of frame */
	glTranslatef(0,0.0,1.4); /* float model in front of user */
	glScalef(obj_scale,obj_scale,obj_scale);
	if (model_gui==mat4(1.0)) 
	{ /* User hasn't touched controls--rotate slowly */
		//glRotatef(nice_timer()*1.3,1.0,0.0,0); /* adjust model axis */
		glRotatef(nice_timer()*10,0.0,0.0,1); /* slowly rotate model */
	}
	glRotatef(90.0,1.0,0,0); /* rotate model to face user */
	glMultMatrixf(&model_gui[0][0]);
	glGetFloatv(GL_MODELVIEW_MATRIX, &modelM[0][0]);
	
	glutPostRedisplay(); //<- for model animation
	
	/* Do model global illumination */
	model.gi(modelM);
}

void my_display(void) {
	glFinish();
	int last_msec=glutGet(GLUT_ELAPSED_TIME);
	
	if (show_sky) 
		glClearColor(0.1,0.3,0.4,0); /* set background color to blue */
	else
		glClearColor(0.0,0.0,0.0,0); /* set background color to black */
	
	if (!world) world=new the_world();
	
/* Do global illumination stuff */
	world->setup();
	
	/* Show the GI stuff */
	if (show_gi) { 
		glutSwapBuffers();
		return;
	}
	
/* Reset matrices after GI stuff */
	int w=glutGet(GLUT_WINDOW_WIDTH), h=glutGet(GLUT_WINDOW_HEIGHT);
	glViewport(0,0,w,h);
	glScissor(0,0,w,h);
	
	glMatrixMode(GL_PROJECTION); /* set up projection matrix */
	glLoadIdentity(); /* reset matrix at start of frame */
	
/* Set up final display camera */
	gluPerspective(90.0, /* <- Y camera field-of-view, in degrees */
		glutGet(GLUT_WINDOW_WIDTH)/(float)glutGet(GLUT_WINDOW_HEIGHT), /* aspect ratio: width over height */
		0.01, /* Near clipping plane depth */
		1000.0 /* Far clipping plane depth */
	);
	oglCameraLookAt(1.0+length(camera));
	
	glMatrixMode(GL_MODELVIEW); /* set up model and view matrix */
	glLoadIdentity(); /* reset matrix at start of frame */
	
	glClear(GL_COLOR_BUFFER_BIT + GL_DEPTH_BUFFER_BIT);
	glEnable(GL_DEPTH_TEST);

/* Draw the world */
	world->render(true);
	
	glutSwapBuffers(); /* copy back buffer to front */
	glutPostRedisplay();
	
/* Statistics */
	glFinish();
	int cur_msec=glutGet(GLUT_ELAPSED_TIME); /* milliseconds since glutInit */
	int msec_elapsed=cur_msec-last_msec;
	printf("%d msec/frame; %.2f fps\n",
		msec_elapsed,1.0e3/msec_elapsed);
	last_msec=cur_msec;
	seconds_per_frame=msec_elapsed*1.0e-3;
}

int main(int argc,char *argv[]) {
	glutInitWindowSize(800,600);
	glutInit(&argc,argv);
	int mode=GLUT_DOUBLE | GLUT_RGBA | GLUT_DEPTH;
        glutInitDisplayMode (mode);
	int win=glutCreateWindow("Interactive Per-Vertex Global Illumination Demo (try the WASD keys)");
	glewInit();
	
	oglCameraInit();
	camera=vec3(0.0,-3.5,2.5);
	
	GLUI *glui = GLUI_Master.create_glui_subwindow(win, GLUI_SUBWINDOW_TOP);
	glui->set_main_gfx_window( win );
	
	new GLUI_Checkbox(glui, "Show GI Array?", &show_gi);
	new GLUI_Spinner(glui, "GI FOV: ", &gi_fov);
	new GLUI_Spinner(glui, "GI Res: 2^", &gi_level);
	new GLUI_Checkbox(glui, "Cosine Incoming", &cos_falloff);
	
	new GLUI_Column(glui); /* for layout */
	new GLUI_Checkbox(glui, "Blue Sky", &show_sky);
	new GLUI_Checkbox(glui, "Ground", &show_ground);
	new GLUI_Checkbox(glui, "Noise Texture", &show_noise);
	new GLUI_Checkbox(glui, "Light Source", &show_light);
	new GLUI_Checkbox(glui, "Light Antialias", &light_antialias);
	
	
	new GLUI_Column(glui); /* for layout */
	new GLUI_Spinner(glui, "Model Reflectance: ", &obj_reflectance);
	new GLUI_Spinner(glui, "Model Size: ", &obj_scale);
	(new GLUI_Spinner( glui, "Light Size:", &lightsize))
		->set_float_limits(0.0, 40.0);
	new GLUI_Checkbox(glui, "Model Self-lit", &gi_self);
	//new GLUI_Spinner(glui, "Walking Speed: ", &walking_speed);
	//new GLUI_Spinner(glui, "model Count: ", &nmodel);
	//new GLUI_Checkbox(glui, "Move camera?", &move_camera);
	
	new GLUI_Column(glui); /* for layout */
	new GLUI_Rotation(glui,"Model",&model_gui[0][0]);
	
	new GLUI_Column(glui);
	new GLUI_Rotation(glui,"Light:",&lightMat[0][0]);
	
	glutDisplayFunc(my_display);
	
	glutMainLoop(); /* processes events */
	return 0;
}
